/*
 *   Copyright 2006-2007 Aaron Seigo <aseigo@kde.org>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Library General Public License as
 *   published by the Free Software Foundation; either version 2, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this program; if not, write to the
 *   Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "querymatch.h"

#include <QPointer>
#include <QVariant>
#include <QStringList>
#include <QIcon>

#include <KDebug>

#include "abstractrunner.h"

namespace Plasma
{

class QueryMatch::Private
{
    public:
        Private(AbstractRunner *r)
            : runner(r),
              type(QueryMatch::ExactMatch),
              enabled(true),
              relevance(.7)
        {
        }

        QPointer<AbstractRunner> runner;
        QueryMatch::Type type;
        QString id;
        QString text;
        QString subtext;
        QIcon icon;
        QVariant data;
        bool enabled;
        qreal relevance;
};


QueryMatch::QueryMatch(AbstractRunner *runner)
    : d(new Private(runner))
{
    d->id = runner->id();
//    kDebug() << "new match created";
}

QueryMatch::~QueryMatch()
{
    delete d;
}

QString QueryMatch::id() const
{
    return d->id;
}

void QueryMatch::setType(Type type)
{
    d->type = type;
}

QueryMatch::Type QueryMatch::type() const
{
    return d->type;
}

void QueryMatch::setRelevance(qreal relevance)
{
    d->relevance = qMax(qreal(0.0), qMin(qreal(1.0), relevance));
}

qreal QueryMatch::relevance() const
{
    return d->relevance;
}

AbstractRunner* QueryMatch::runner() const
{
    return d->runner;
}

void QueryMatch::setText(const QString& text)
{
    d->text = text;
}

void QueryMatch::setSubtext(const QString& subtext)
{
    d->subtext = subtext;
}

void QueryMatch::setData(const QVariant& data)
{
    d->data = data;
    d->id = d->runner->id() + '_' + data.toString();
}

void QueryMatch::setIcon(const QIcon& icon)
{
    d->icon = icon;
}

QVariant QueryMatch::data() const
{
    return d->data;
}

QString QueryMatch::text() const
{
    return d->text;
}

QString QueryMatch::subtext() const
{
    return d->subtext;
}

QIcon QueryMatch::icon() const
{
    return d->icon;
}

void QueryMatch::setEnabled( bool enabled )
{
    d->enabled = enabled;
}

bool QueryMatch::isEnabled() const
{
  return d->enabled;
}

bool QueryMatch::operator<(const QueryMatch& other) const
{
    return d->relevance < other.d->relevance;
}

void QueryMatch::run(const RunnerContext *context) const
{
    Q_ASSERT(context);

    //kDebug() << "we run the term" << context->query() << "whose type is" << context->mimetype();
    if (d->runner) {
        //TODO: this could be dangerous if the runner is deleted behind our backs.
        d->runner->run(context, this);
    }
}

} // Plasma namespace

